-------------------------------------------------------------------------------
--! @file       PCCTRL.vhd
--! @author     Ahmed Ferozpuri
--! @brief      PRIMATEs-GIBBON Cipher Controller
-------------------------------------------------------------------------------
LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
USE work.PRIMATEsPKG.ALL;

ENTITY PCCTRL IS
	PORT(	
		  --GMU INTERFACE SIGNALS
        clk                       : IN  std_logic;
        rst                       : IN  std_logic;

        key_ready                 : OUT  std_logic;
		key_valid                 : IN  std_logic;
        key_update                : IN std_logic;
		bdi_ready                 : OUT  std_logic;
		bdi_valid                 : IN  std_logic;
		bdi_type                  : IN  std_logic_vector(3                -1 downto 0);
		
        decrypt                   : IN  std_logic;
        bdi_eot                   : IN  std_logic;
        bdi_eoi                   : IN  std_logic;
        bdi_size                  : IN  std_logic_vector(3              			    downto 0);

        bdo_sel			          : OUT std_logic;
		bdo_ready                 : IN  std_logic;
        bdo_valid                 : OUT std_logic;
        msg_auth_done             : OUT std_logic;
		  
		--USER DEFINED SIGNALS (TO DATAPATH)
		init                      : OUT  std_logic;
		run                       : OUT  std_logic;
		mode                      : OUT  std_logic;
		lastBlockFull             : OUT  std_logic;
		selV                      : OUT  std_logic_vector(2 								 DOWNTO 0);
		pn 						  : OUT  std_logic_vector(1 								 DOWNTO 0);
		done	                  : IN std_logic;
		
		--SIPO/PISO SIGNALS
		npub_en                   : OUT  std_logic;
		etag_en                   : OUT  std_logic;
		tag_en                    : OUT  std_logic;
		tag_ld                    : OUT  std_logic
		);
END PCCTRL;

ARCHITECTURE behavioral OF PCCTRL IS

	TYPE gStates IS (	IDLE, ACTIVATE_KEY, LOAD_NPUB, LOAD_INIT, RUN_INIT, LOAD_POST_INIT,
						LOAD_DATA,
						RUN_PRE_AD,
						RUN_BLOCK,
						LOAD_EMPTY_MC,
						LOAD_PRE_TAG, RUN_PRE_TAG,
						GENVER_TAG, WRITE_TAG);
						  
	SIGNAL gps, gns : gStates;
	SIGNAL bdi_eot_r, bdi_eot_l, bdi_eoi_r, bdi_eoi_l, decrypt_r, decrypt_l, bdi_type_l : std_logic;
	SIGNAL blockFull, blockEmpty : std_logic;
	SIGNAL pre_ad_compl, pre_ad_compl_s, pre_ad_compl_r : std_logic;
	SIGNAL bdi_type_r : std_logic_vector(3			-1 downto 0);
	
BEGIN
	
	blockFull <= '1' WHEN bdi_size = "0101" ELSE '0';	
	lastBlockFull <= bdi_eot and blockFull and pre_ad_compl;
	blockEmpty <= '1' WHEN bdi_size = "0000" ELSE '0';
	mode <= decrypt;
	
	GIBBON_STATE: PROCESS (clk, rst)
	BEGIN
		IF rst = '1' THEN
			gps <= IDLE;
			bdi_eoi_r <= '0';
			bdi_eot_r <= '0';
			decrypt_r <= '0';
			pre_ad_compl <= '0';
			bdi_type_r <= (others => '0');
		ELSIF rising_edge(clk) THEN
			gps <= gns;
			
			IF (pre_ad_compl_r = '1') THEN
				pre_ad_compl <= '0';
			ELSIF (pre_ad_compl_s = '1') THEN
				pre_ad_compl <= '1';
			END IF;
			
			IF (bdi_eoi_l = '1') THEN 
				bdi_eoi_r <= bdi_eoi;
			END IF;
			
			IF (bdi_eot_l = '1') THEN 
				bdi_eot_r <= bdi_eot;
			END IF;
			
			IF (bdi_type_l = '1') THEN
				bdi_type_r <= bdi_type;
			END IF;
			
			IF (decrypt_l = '1') THEN
				decrypt_r <= decrypt;
			END IF;
		END IF;
	END PROCESS GIBBON_STATE;

	GIBBON_OUTPUT: PROCESS (rst, gps, key_valid, key_update, bdi_valid, done,
							bdi_eot, bdi_eoi, bdo_ready, decrypt, decrypt_r, bdi_eot_r, bdi_eoi_r,
							bdi_type, pre_ad_compl, bdi_type_r, blockEmpty)
		BEGIN
		IF rst = '1' THEN
			gns <= IDLE;
		ELSE
			init <= '0';
			run <= '0';
			bdo_valid <= '0';
			bdi_ready <= '0';
			pn <= (OTHERS => '0');
			selV <= (OTHERS => '0');
			msg_auth_done <= '0';
			gns <= gps;
			key_ready <= '0';
			bdi_ready <= '0';
			npub_en <= '0';
			etag_en <= '0';
			tag_en <= '0';
			tag_ld <= '0';
			bdo_sel <= '0';
			bdi_eoi_l <= '0';
			bdi_eot_l <= '0';
			bdi_type_l <= '0';
			decrypt_l <= '0';
			pre_ad_compl_s <= '0';
			pre_ad_compl_r <= '0';
			
			CASE gps IS
				WHEN IDLE =>
					pre_ad_compl_r <= '1';
					IF (bdi_valid = '1') THEN
						IF (key_update = '1') THEN
							gns <= ACTIVATE_KEY;
						ELSE
							gns <= LOAD_NPUB;
						END IF;
					END IF;
				WHEN ACTIVATE_KEY =>
					key_ready <= '1';
					IF (key_valid = '1') THEN
						gns <= LOAD_NPUB;
					END IF;
				WHEN LOAD_NPUB =>
					decrypt_l <= '1';
					bdi_eoi_l <= '1';
					bdi_type_l <= '1';
					IF (bdi_valid = '1') THEN
						bdi_ready <= '1';
						npub_en <= '1';
					END IF;
					IF (bdi_eot = '1') THEN
						gns <= LOAD_INIT;
					END IF;
				WHEN LOAD_INIT =>
					init <= '1';
					selV <= "000";
					gns <= RUN_INIT;
				WHEN RUN_INIT =>
					run <= '1';
					IF (done = '1') THEN
						gns <= LOAD_POST_INIT;
					END IF;
				WHEN LOAD_POST_INIT =>
					init <= '1';
					selV <= "001";
					gns <= LOAD_DATA;
				WHEN LOAD_DATA =>
					bdi_eot_l <= '1';
					bdi_eoi_l <= NOT bdi_eoi_r;
					bdi_type_l <= NOT bdi_eoi_r;
					decrypt_l <= '1';
					IF (bdi_eoi_r = '1' AND bdi_type_r = "110") THEN
						--bdi_ready <= '1';
						selV <= "110";
						pn <= "10";
						init <= '1';
						gns <= RUN_BLOCK;
					ELSIF (bdi_valid = '1') THEN
						bdi_ready <= '1';
						init <= '1';
						IF (bdi_type = "000") THEN
							IF (pre_ad_compl = '0') THEN
								pre_ad_compl_s <= '1';
								selV <= "110";
								pn <= "01";
								bdi_ready <= '0';
								gns <= RUN_PRE_AD;
							ELSE
								selV <= "100";
								IF (bdi_eot /= '1') THEN
									pn <= "01";
								ELSE
									pn <= "10";
								END IF;
								IF (blockEmpty = '1') THEN
									gns <= LOAD_DATA;
								ELSE
									gns <= RUN_BLOCK;
								END IF;
							END IF;
						ELSIF (bdo_ready = '1') THEN
							IF (pre_ad_compl = '0') THEN
								pre_ad_compl_s <= '1';
								selV <= "110";
								pn <= "10";
								init <= '1';
								bdi_ready <= '0';
								bdi_type_l <= '0';
								bdi_eoi_l <= '0';
								gns <= RUN_BLOCK;
							ELSE
								bdo_valid <= '1';
								bdo_sel <= '1';
								pn <= "10";
								IF (decrypt_r = '1') THEN
									selV <= "010"; --M/C || Vc
								ELSE
									selV <= "100";
								END IF;
								gns <= RUN_BLOCK;
							END IF;
						END IF;
					END IF;
				WHEN RUN_BLOCK =>
					run <= '1';
					IF (done = '1') THEN
						gns <= LOAD_DATA;
						IF (bdi_type_r = "010" and bdi_eot_r = '1') THEN
							gns <= LOAD_PRE_TAG;
						ELSIF ((bdi_type_r = "110" OR bdi_type_r = "000") and bdi_eoi_r = '1') THEN
							bdi_type_l <= '1';
							gns <= LOAD_EMPTY_MC;
						END IF;
					END IF;
				WHEN LOAD_EMPTY_MC =>
					bdi_ready <= '1';
					selV <= "100";
					pn <= "10";
					init <= '1';
					gns <= RUN_BLOCK;
				WHEN RUN_PRE_AD =>
					run <= '1';
					IF (done = '1') THEN
						gns <= LOAD_DATA;
					END IF;
				WHEN LOAD_PRE_TAG =>
					init <= '1';
					selV <= "001";
					gns <= RUN_PRE_TAG;
				WHEN RUN_PRE_TAG =>
					run <= '1';
					IF (done = '1') THEN
						gns <= GENVER_TAG;
					END IF;
				WHEN GENVER_TAG =>
					IF (bdi_valid = '1' AND decrypt_r = '1') THEN
						bdi_ready <= '1';
						etag_en <= '1';
						IF (bdi_eot = '1') THEN --last etag block
							gns <= WRITE_TAG;
						END IF;
					ELSIF (bdo_ready = '1') THEN
						tag_ld <= '1';
						gns <= WRITE_TAG;
					END IF;
				when WRITE_TAG =>
					IF (decrypt_r = '1') THEN
						msg_auth_done <= '1';
						gns <= IDLE;
					ELSIF (bdo_ready = '1') THEN
						tag_en <= '1';
						bdo_valid <= '1';
					ELSE
						gns <= IDLE;
					END IF;
			END CASE;
		END IF;
	END PROCESS GIBBON_OUTPUT;

END behavioral;